#include "network.h"
#include "sigint.h"
#include "main.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <error.h>
#include <errno.h>
#include <sys/errno.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>

extern "C" {
#include <mcrypt.h>
}

network::network()
 {
  len = 0;
  buf = 0;
  fd = 1;
  if (fd < 0) throw couldNotOpen();
  localSecure = 0;
  remoteSecure = 0;
  encryption = -1;
  handshook = false;
  localIV = 0;
  remoteIV = 0;
  master = false;
  localCryptor = 0;
  remoteCryptor = 0;
  closeHandle = false;
 }
 
network::network(char *host, int p)
 {
  struct hostent *he;
  struct sockaddr_in peer;

  if (verbose) 
   printf("NET: Making connection to %s:%i.\n", host, p);

  he = gethostbyname(host);
  if (!he) throw couldNotOpen();

  fd = socket(PF_INET, SOCK_STREAM, 0);
  if (fd < 0) throw couldNotOpen();

  peer.sin_family = AF_INET;
  peer.sin_port = htons(p);
  peer.sin_addr = *(struct in_addr*) he->h_addr;

  if (connect(fd, (struct sockaddr*) &peer, sizeof(peer)) < 0) throw couldNotOpen();

  if (verbose) 
   printf("NET: Connected.\n");

  closeHandle = true;
 }

network::network(int p)
 {
  struct sockaddr_in sa, peer;
  fd_set fs;
  int fd_listen;
  unsigned int l;

  if (verbose) 
   printf("NET: Waiting for connection on port %i.\n", p);

  fd_listen = socket(PF_INET, SOCK_STREAM, 0);
  if (fd_listen < 0) throw couldNotOpen();
      
  sa.sin_family = AF_INET;
  sa.sin_port = htons(p);
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  
  if (bind(fd_listen, (struct sockaddr*) &sa, sizeof(sa)) < 0) throw couldNotOpen();
  if (listen(fd_listen, 1) < 0) throw couldNotOpen();

  FD_ZERO(&fs);
  FD_SET(fd_listen, &fs);
  select(FD_SETSIZE, &fs, 0, 0, 0);
  
  if (sigIntOccured) 
   {
    sigIntOccured = false;
    close(fd_listen);
    throw userBreak();
   }
    
  fd = accept(fd_listen, (sockaddr*) &peer, &l);
  if (fd < 0) throw couldNotOpen();
  
  close(fd_listen);
 
  if (verbose) 
   printf("NET: Connected.\n");
  
  closeHandle = true;  
 }
 
network::~network()
 {
  if (closeHandle && (fd >= 0)) close(fd);
  if (buf) delete [] buf;
  
  if (localIV) delete [] localIV;
  if (remoteIV) delete [] remoteIV;
  
  if (localCryptor) end_mcrypt_cfb(localCryptor);
  if (remoteCryptor) end_mcrypt_cfb(remoteCryptor);
 }
   
void network::read()
 {
  unsigned char b;
  int r, l;
  char *d;

  if (!handshook) handshake();
  
  len = readInteger();

  if (buf) delete [] buf;
  d = buf = new char [len];
  l = len;
  while(l)
   {
    r = ::read(fd, d, l);
    if (r <= 0) throw couldNotRead();
    l -= r; d += r;
   }
   
  mdecrypt_cfb(localCryptor, buf, len);
 }
 
int network::getBufLen()
 {
  return len;
 }

char *network::getBuf()
 {
  return buf;
 }

void network::write(char *d, int s)
 {
  unsigned char b;

  if (!handshook) handshake();
  writeInteger(s);

  mcrypt_cfb(remoteCryptor, d, s);
  if (::write(fd, d, s) != s) throw couldNotWrite();
 }
 
network::operator int()
 {
  return fd;
 }

void network::setLocalSecure(char *s)
 {
  if (localSecure) delete [] localSecure;
  localSecure = strcpy(new char[strlen(s)+1], s);
 }
 
void network::setRemoteSecure(char *s)
 {
  if (remoteSecure) delete [] remoteSecure;
  remoteSecure = strcpy(new char[strlen(s)+1], s);
 }
 
void network::setEncryption(int e)
 {
  if ((e < -1) || (e > 19) || (e == 5)) throw unknownEncryption();
  encryption = e;
 }

void network::handshake()
 {
  char *t;
  int iv;
  if (handshook) return;

  try
   {
    electMaster();
   }    
  catch (couldNotRead)  { throw handshakeFailure(); }    
  catch (couldNotWrite) { throw handshakeFailure(); }
    
   
  if (verbose) 
   printf("NET: Election was held, we are %s.\n", master ? "master" : "client");
  
  try
   {
    if (master)
     {
      if (readInteger() != encryption) throw encryptionInconsistent();
      writeInteger(encryption);
     }    
    else
     {
      writeInteger(encryption);
      if (readInteger() != encryption) throw encryptionInconsistent();
     } 
   }
  catch (couldNotRead)  { throw encryptionInconsistent(); } 
  catch (couldNotWrite) { throw encryptionInconsistent(); } 
     
  if (verbose) 
   printf("NET: Encryption is ok: %s\n", get_algorithms_name(encryption));
  
  try 
   {
    if (master)
     {
      createLocalIV();
      readRemoteIV();
     }
    else
     {
      readRemoteIV();
      createLocalIV();
     }
   }
  catch (couldNotRead) { throw handshakeFailure(); }
  catch (couldNotWrite) { throw handshakeFailure(); }

  if (verbose) 
   printf("NET: IVs exchanged.\n");

  localCryptor  = init_mcrypt_cfb(encryption, localSecure, strlen(localSecure), localIV);
  if (localCryptor == -1) throw cryptError();
  
  remoteCryptor = init_mcrypt_cfb(encryption, remoteSecure, strlen(remoteSecure), remoteIV);
  if (remoteCryptor == -1) throw cryptError();
   
  try
   {  
    if (master)
     {
      authClient();
      authServer();
     }
    else
     {
      authServer();
      authClient();
     }
   }
  catch (couldNotRead)  { throw authFailure(); }
  catch (couldNotWrite) { throw authFailure(); }

  if (verbose) 
   printf("NET: Authorized.\n");
   
  handshook = true;
 }

void network::electMaster()
 {
  int r1, r2;
 
  do
   {
    r1 = rand();
    writeInteger(r1);
    r2 = readInteger();
   }
  while (r1 == r2);

  master = (r1 > r2) ? true : false;
 }

void network::authServer()
 {
  int v;
  v = readInteger();
  mdecrypt_cfb(localCryptor, &v, sizeof(v));
  mcrypt_cfb(remoteCryptor, &v, sizeof(v));
  writeInteger(v);
 }

void network::authClient()
 {
  int v, v2, v3;

  v2 = v = rand();
  mcrypt_cfb(remoteCryptor, &v2, sizeof(v2));
  writeInteger(v2);
  v3 = readInteger();
  mdecrypt_cfb(localCryptor, &v3, sizeof(v3));
  
  if (v != v3) throw authFailure();
 }
 
void network::createLocalIV()
 {
  char *buf = new char[get_block_size(encryption)];
  
  for (int i = 0; i < get_block_size(encryption); i++)
   {
    buf[i] = (char) rand();
    if (::write(fd, &buf[i], sizeof(char)) != sizeof(char))
     throw couldNotWrite();
   }    
  
  localIV = (void*) buf;
 } 

void network::readRemoteIV()
 {
  char *buf = new char[get_block_size(encryption)];

  for (int i = 0; i < get_block_size(encryption); i++)
   {
    if (::read(fd, &buf[i], sizeof(char)) != sizeof(char)) 
     throw couldNotRead();
   }     

  remoteIV = (void *) buf;
 }
 
 
void network::writeInteger(int i)
 {
  unsigned char b;
  
  b = (unsigned char) ((i >> 24) & 0xFF);
  if (::write(fd, &b, 1) != 1) throw couldNotWrite();
  b = (unsigned char) ((i >> 16) & 0xFF);
  if (::write(fd, &b, 1) != 1) throw couldNotWrite();
  b = (unsigned char) ((i >> 8) & 0xFF);
  if (::write(fd, &b, 1) != 1) throw couldNotWrite();
  b = (unsigned char) (i & 0xFF);
  if (::write(fd, &b, 1) != 1) throw couldNotWrite();
 }

int network::readInteger()
 {
  unsigned char b;
  int i;

  if (::read(fd, &b, sizeof(b)) != sizeof(b)) throw couldNotRead();
  i = (int) b << 24;
  if (::read(fd, &b, sizeof(b)) != sizeof(b)) throw couldNotRead();
  i |= (int) b << 16;
  if (::read(fd, &b, sizeof(b)) != sizeof(b)) throw couldNotRead();
  i |= (int) b << 8;
  if (::read(fd, &b, sizeof(b)) != sizeof(b)) throw couldNotRead();
  i |= (int) b;

  return i;
 }
 
